"""Validation functionality."""
import importlib
import inspect
import json
import logging
import sys

from jsonschema import validate
from jsonschema.exceptions import ValidationError

from django.conf import settings
from django.http.response import HttpResponse, HttpResponseNotAllowed, JsonResponse
from django.views.decorators.csrf import csrf_exempt

from . exceptions import FatalException, UnsupportedMediaTypeException

logger = logging.getLogger(__name__)


class ContentType:
    """Represents http content types."""
    JSON = 'application/json'

    def __init__(self):
        """Initialisation function."""
        pass


class Endpoint:
    """
    Endpoint that represents one url in the service. Each endpoint
    contains Actions which represent a request method that the endpoint
    supports.
    """

    url = None
    request_method_mapping = None

    def __init__(self, url):
        """Initialisation function."""
        self.url = url
        self.request_method_mapping = {}

    def parse_regex(self, regex_dict):
        """
        Replace dynamic template in url with corresponding regex for a dynamic value
        :param regex_dict: dictionary of dynamic id names to regex
        e.g. {'dynamic_id': '(?P<dynamic_id>[a-zA-Z]+)'}
        """

        for regex_key, regex in regex_dict.items():
            string_to_replace = "{%s}" % regex_key
            self.url = self.url.replace(string_to_replace, regex)

    def add_action(self, request_method, action):
        """Add an action mapping for the given request method type.
        :param request_method: http method type to map the action to.
        :param action: the action to map to the request.
        :returns: returns nothing.
        """

        self.request_method_mapping[request_method] = action

    @csrf_exempt
    def serve(self, request, **dynamic_values):
        """Serve the request to the current endpoint. The validation and response
        that is returned depends on the incoming request http method type.
        :param request: incoming http request that must be served correctly.
        :param dynamic_values: kwargs of dynamic id names against actual value to substitute into url
         e.g. {'dynamic_id': 'aBc'}
        :returns: returns the HttpResponse, content of which is created by the target function.
        """

        if request.method in self.request_method_mapping:
            action = self.request_method_mapping[request.method]
            response = _validate_api(request, action, dynamic_values)
        else:
            response = HttpResponseNotAllowed(self.request_method_mapping.keys())

        if isinstance(response, HttpResponse):
            return response
        else:
            return HttpResponse(response)


class Action:
    """
    Maps out the api definition associated with the parent Endpoint.
    One of these will be created per http request method type.
    """

    example = None
    schema = None
    target = None
    query_parameter_checks = None
    resp_content_type = None
    requ_content_type = None
    regex = None

    def __init__(self):
        """Initialisation function."""
        pass


def _validate_query_params(params, checks):
    """
    Function to validate HTTP GET request params. If there are checks to be
    performed then they will be; these will be items such as length and type
    checks defined in the definition file.
    :param params: incoming request parameters.
    :param checks: dict of param to rule to validate with.
    :raises ValidationError: raised when any query parameter fails any
        of its checks defined in the checks param.
    :returns: true if validated, otherwise raises an exception when fails.
    """

    # If validation checks, check the params. If not, pass.
    if checks:
        for param in checks:
            # If the expected param is in the query.
            if param in params:
                for check, rule in checks[param].items():
                    if rule is not None:
                        error_message = 'QueryParam [%s] failed validation check [%s]:[%s]' % (param, check, rule)
                        if check == 'minLength':
                            if len(params.get(param)) < rule:
                                raise ValidationError(error_message)
                        elif check == 'maxLength':
                            if len(params.get(param)) > rule:
                                raise ValidationError(error_message)
                        elif check == 'type':
                            if rule == 'number':
                                try:
                                    float(params.get(param))
                                except ValueError:
                                    raise ValidationError(error_message)

            # If the require param isn't in the query.
            elif checks[param]['required'] is True:
                raise ValidationError('QueryParam [%s] failed validation check [Required]:[True]' % param)

    return True


def _generate_example(action):
    """
    This is used by both GET and POST when returning an example
    """
    # The original method of generating straight from the example is bad
    # because v2 parser now has an object, which also allows us to do the
    # headers correctly

    ret_data = action.example
    # FIXME: not sure about this content thing
    if action.resp_content_type == "application/json":
        ret_data = json.dumps(action.example)

    return HttpResponse(ret_data, content_type=action.resp_content_type)


def _validate_api(request, action, dynamic_values=None):
    """
    Validate APIs content.
    :param request: incoming http request.
    :param action: action object containing data used to validate
        and serve the request.
    :param dynamic_values: dict of dynamic id names against actual values to substitute into url
     e.g. {'dynamic_id': 'aBc'}
    :returns: returns the HttpResponse generated by the action target.
    """

    if action.query_parameter_checks:
        # Following raises exception on fail or passes through.
        _validate_query_params(request.GET, action.query_parameter_checks)

    error_response = None

    if request.body:
        error_response = _validate_body(request, action)

    if error_response:
        response = error_response
    else:
        if action.target:
            if dynamic_values:
                # If there was a dynamic value, pass it through
                response = action.target(request, **dynamic_values)
            else:
                response = action.target(request)
        else:
            response = _generate_example(action)

    if not isinstance(response, HttpResponse):
        # As we weren't given a HttpResponse, we need to create one
        # and handle the data correctly.
        if action.resp_content_type == ContentType.JSON:
            response = HttpResponse(json.dumps(response), content_type="application/json")
        else:
            # FIXME: write more types in here
            raise Exception("Unsuported response content type - contact @jmons for future feature request")

    return response


def _validate_body(request, action):
    error_response = None
    content_type_matched = False

    try:
        # Grab the content-type coming in from the request
        if "headers" in request.META:
            request_content_type = request.META["headers"]["content-type"]

        else:
            request_content_type = request.META["CONTENT_TYPE"]
    except Exception:
        # couldn't find the content-type header so error
        error_response = _validation_error_handler(UnsupportedMediaTypeException("Missing Content Type for this request"))
        return error_response

    if not request_content_type:
        # content-type is empty so error
        error_response = _validation_error_handler(UnsupportedMediaTypeException("Missing Content Type for this request"))
        return error_response

    # Parse the content type coming in (in case there are multiple optional entries)
    request_content_type = request_content_type.replace(' ', '').split(';')[0]

    # Set the actual content_type we are using in this request
    action.requ_content_type = request_content_type

    # Check the schema had content-types defined
    if hasattr(action, 'request_content_type_options'):

        for x in action.request_content_type_options:
            # Check if the incoming content-type matches the allowed type in the schema and is JSON type
            if x == request_content_type == str(ContentType.JSON):
                content_type_matched = True

                # If the expected request body is JSON, we need to load it.
                if action.request_options[request_content_type]["schema"]:
                    # If there is any schema, we'll validate it.
                    try:
                        data = json.loads(request.body.decode('utf-8'))
                        validate(data, action.request_options[request_content_type]["schema"])
                    except Exception as e:
                        # Check the value is in settings, and that it is not None
                        if hasattr(settings,
                                   'RAMLWRAP_VALIDATION_ERROR_HANDLER') and settings.RAMLWRAP_VALIDATION_ERROR_HANDLER:
                            error_response = _call_custom_handler(e, request, action)
                        else:
                            error_response = _validation_error_handler(e)
                else:
                    # Otherwise just load it (no validation as no schema).
                    data = json.loads(request.body.decode('utf-8'))
                break

            # Incoming content type wasn't json but it does match one of the options in the raml so just decode it as is
            elif x == request_content_type:
                content_type_matched = True
                try:
                    data = request.body.decode('utf-8')
                except UnicodeDecodeError:
                    # Just send the body if it cannot be decoded
                    data = request.body
                break

    else:
        # There were no content type options in the schema so just load the data
        content_type_matched = True
        try:
            data = request.body.decode('utf-8')
        except UnicodeDecodeError:
            # Just send the body if it cannot be decoded
            data = request.body

    if not content_type_matched:
        error_response = _validation_error_handler(ValidationError("Invalid Content Type for this request: {}".format(request_content_type), validator="invalid"))

    if not error_response:
        request.validated_data = data

    return error_response


def _validation_error_handler(e):
    """
    Default validation handler for when a ValidationError occurs.
    This behaviour can be overriden in the settings file.
    :param e: exception raised that must be handled.
    :returns: HttpResponse with status depending on the error.
        ValidationError will return a 422 with json info on the cause.
        Otherwise a FatalException is raised.
    """

    if isinstance(e, ValidationError):
        message = 'Validation failed. {}'.format(e.message)
        error_response = {
            'message': message,
            'code': e.validator
        }
        logger.info(message)
        error_resp = JsonResponse(error_response, status=422)

    elif isinstance(e, UnsupportedMediaTypeException):
        error_response = {'message': e.message}
        error_resp = JsonResponse(error_response, status=415)

    else:
        raise FatalException('Malformed JSON in the request.', 400)

    return error_resp


def _call_custom_handler(e, request, action):
    """
    Dynamically import and call the custom validation error handler
    defined by the user in the django settings file.
    :param e: exception raised that must be handled.
    :param request: incoming http request that must be served correctly.
    :param action: action object containing data used to validate and serve the request.
    :returns: response returned from the custom handler, given the exception.
    """

    handler_full_path = settings.RAMLWRAP_VALIDATION_ERROR_HANDLER
    handler_method = handler_full_path.split('.')[-1]
    handler_class_path = '.'.join(handler_full_path.split('.')[0:-1])

    handler = getattr(importlib.import_module(handler_class_path), handler_method)

    if _num_arguments_to_pass(handler) == 3:
        return handler(e, request, action)
    else:
        # Handle old versions that still only accept the exception
        return handler(e)


def _num_arguments_to_pass(handler):
    if sys.version_info[0] < 3:
        # Python 2
        args = inspect.getargspec(handler).args
        return len(args)
    else:
        # Python 3
        signature = inspect.signature(handler)
        return len(signature.parameters)



from django.http import HttpResponse
from django.shortcuts import render
from collections import OrderedDict

import yaml, copy

from .utils.yaml_include_loader import Loader

try:
    # This import fails for django 1.9
    from django.views import View
except Exception as error:
    print("failed to [from django.views import View] {}".format(error))
    from django.views.generic.base import View


class RamlDoc(View):
    """
    As a view system, this should be called with the raml_file set. If desired
    then replacing the template is also possible with the variable `template`.
    """

    raml_file = None
    # FIXME: make this inside ramlwrap and fix setup.py to have fixtures
    template = 'ramlwrap_default_main.html'

    def get(self, request):
        # WARNING multi return function (view switching logic)
        context = self._parse_endpoints(request)

        if "type" in request.GET:
            if request.GET['type'] in ("single_api", "schema"):
                target_endpoint = request.GET['entry']
                endpoints = context['endpoints']
                for suspect in endpoints:
                    if suspect.url == target_endpoint:
                        endpoints = [ suspect ]
                        context['endpoints'] = endpoints

            if request.GET['type'] == "schema":
                if request.GET['schema_direction'] == "request":
                    # FIXME ok this doesnt work. Writing here- need to have a better string to work out what you want.
                    return HttpResponse(context['endpoints'][0].request_schema)
                else:
                    raise Exception("Not implemneted yet - response schema")

        return render(request, self.template, context)

    def _parse_endpoints(self, request):

        # Read Raml file
        file = open(self.raml_file)
        loaded_yaml = yaml.load(file, Loader=Loader)  # This loader has the !include directive
        file.close()

        # Parse raml file and generate the tree
        item_queue = [] # Queue
        endpoints  = [] # Endpoint list

        # Parse root and queue its children
        root_element = {
            "node": loaded_yaml,
            "path": ""
        }
        _parse_child(root_element, endpoints, item_queue, True)

        # Keep parsing children queue until empty
        while item_queue:
            item = item_queue.pop(0)
            _parse_child(item, endpoints, item_queue)

        # Build object with parsed data
        context = {
            "endpoints" : endpoints
        }

        # Root attributes
        for tag in ["title", "description", "version", "mediaType", "baseUri"]:
            if tag in loaded_yaml:
                context[tag] = loaded_yaml[tag]
            else:
                context[tag] = None

        # Nested attributes
        for tag in ["documentation", "traits", "securitySchemes"]:
            context[tag] = {}
            if tag in loaded_yaml:
                for key in loaded_yaml[tag][0]:
                    context[tag][key] = loaded_yaml[tag][0][key]

        # Return the data to the template
        return context


# FIXME delete this before committing. Holder for Gateway
def noscript(request):
    return HttpResponse("")


class Endpoint():
    url          = ""
    description  = ""
    request_type = ""
    methods      = None
    level        = 0

    def __init__(self):
        self.methods = []


class Method():
    method_type = ""
    displayName = ""
    description = ""

    request_content_type  = None
    request_schema        = None
    request_schema_original = None
    request_example       = None
    request_examples      = None

    response_content_type = None
    response_schema       = None
    response_example      = None
    response_examples     = None
    response_description  = None

    responses = None
    examples = None

    def __init__(self):
        self.responses = []
        self.examples = []


class Response:
    content_type = None
    schema = None
    schema_original = None
    examples = None
    description = None
    status_code = None

    def __init__(self, content_type=None, schema=None, schema_original=None, description=None, status_code=None):
        self.content_type = content_type
        self.schema = schema
        self.schema_original = schema_original
        self.examples = []
        self.description = description
        self.status_code = status_code


class Example:
    title = None
    body = None

    def __init__(self, title=None, body=None):
        self.title = title
        self.body = body


def _parse_child(resource, endpoints, item_queue, rootnode=False):
    """ Only set root = True on the root of the raml file """
    # look for endpoints etc. This is subtly different to the tree parser in the
    # main ramlwrap usage, although consider a refactor to merge these two into
    # the same. (i.e. the other one takes a function map, where as this doesn't)

    node = resource['node']
    path = resource['path']

    if rootnode:
        level = -2
    else:
        level = resource['level']

    # Storing endpoints in temporary array to preserve order
    child_item_array = []

    # Parse children endpoints
    for key in node:
        if key.startswith("/"):
            item = {
                "node": node[key],
                "path": "%s%s" % (path, key),
                "level": level + 1
            }
            child_item_array.insert(0, item)

    for i in child_item_array:
        item_queue.insert(0, i)

    # Skip rootnode
    if rootnode:
        return

    # Skip empty nodes
    if "displayName" not in node:
        return

    # Parse endpoint data
    current_endpoint = Endpoint()
    current_endpoint.url = path
    current_endpoint.level = level
    endpoints.append(current_endpoint)

    # Endpoint name
    if "displayName" in node:
        current_endpoint.displayName = node["displayName"]

    # Endpoint description
    if "description" in node:
        current_endpoint.description = node["description"]

    # Endpoint methods
    for method_type in ["get", "post", "put", "patch", "delete"]:

        if method_type in node:
            method_data = node[method_type]

            m = Method()
            m.method_type = method_type
            current_endpoint.methods.append(m)

            # Method description
            if "description" in method_data:
                m.description = method_data['description']

            # Request
            if "body" in method_data:
                # if body, look for content type : if not there maybe not valid raml?
                # FIXME: this may be a bug requiring a try/catch - need more real world example ramls
                m.request_content_type = next(iter(method_data['body']))

                # Request schema
                if "schema" in method_data['body'][m.request_content_type]:
                    m.request_schema_original = method_data['body'][m.request_content_type]['schema']
                    m.request_schema = _parse_schema_definitions(copy.deepcopy(m.request_schema_original))

                # Request example
                if "example" in method_data['body'][m.request_content_type]:
                    m.request_example = method_data['body'][m.request_content_type]['example']
                    m.examples.append(Example(body=method_data['body'][m.request_content_type]['example']))

                # Request examples
                if "examples" in method_data['body'][m.request_content_type]:
                    m.request_examples = method_data['body'][m.request_content_type]['examples']
                    # New examples object
                    for example in method_data['body'][m.request_content_type]['examples']:
                        m.examples.append(
                            Example(title=example, body=method_data['body'][m.request_content_type]['examples'][example]))

            # Response
            if 'responses' in method_data and method_data['responses']:
                for status_code in method_data['responses']:
                    _parse_response(m, method_data, status_code)


def _parse_response(m, method_data, status_code):
    response_obj = Response(status_code=status_code)

    response = method_data['responses'][status_code]

    if response:
        for response_attr in response:
            if response_attr == "description":
                response_obj.description = response['description']
            elif response_attr == "body":
                # not sure if this can fail and be valid raml?
                response_obj.content_type = next(iter(response['body']))
                if response['body'][response_obj.content_type]:
                    if "schema" in response['body'][response_obj.content_type]:
                        response_obj.schema_original = response['body'][response_obj.content_type][
                            'schema']
                        response_obj.schema = _parse_schema_definitions(
                            copy.deepcopy(response_obj.schema_original))
                    if "example" in response['body'][response_obj.content_type]:
                        response_obj.examples.append(Example(body=response['body'][response_obj.content_type]['example']))
                        # For backward compatibility
                        if status_code == 200:
                            m.response_example = response['body'][response_obj.content_type]['example']
                    if "examples" in response['body'][response_obj.content_type]:
                        for example in response['body'][response_obj.content_type]['examples']:
                            response_obj.examples.append(
                                Example(title=example, body=response['body'][response_obj.content_type]['examples'][example]))
                        # For backward compatibility
                        if status_code == 200:
                            m.response_example = response['body'][response_obj.content_type]['examples']

    # For backward compatibility, store 200 responses in specific fields
    if status_code == 200:
        m.response_content_type = response_obj.content_type
        m.response_description = response_obj.description
        m.response_schema_original = response_obj.schema_original
        m.response_schema = response_obj.schema

    m.responses.append(response_obj)


def _parse_schema_definitions(schema):

    # If the schema has definitions, replace definitions references with the definition data
    if "definitions" in schema:
        definitions = schema['definitions']

        # Parse definitions in definition properties
        for key in definitions:
            if "properties" in definitions[key]:
                definitions[key]['properties'] = _parse_properties_definitions(definitions[key]['properties'], definitions)

        # Parse definitions in properties
        if "properties" in schema:
            schema['properties'] = _parse_properties_definitions(schema['properties'], definitions)

    return schema


def _parse_properties_definitions(properties, definitions):

    for key in properties:

        # If a property has a $ref definition reference, replace it with the definition
        if "$ref" in properties[key]:
            definition = properties[key]['$ref'].replace("#/definitions/", "")
            if definition in definitions:
                properties[key] = definitions[definition]

        elif "type" in properties[key]:

            # If the property is an object, parse its properties for definitions recursively
            if properties[key]['type'] == "object" and "properties" in properties[key]:
                properties[key]['properties'] = _parse_properties_definitions(properties[key]['properties'], definitions)

            # If the property is an array with a $ref definition reference, replace it with the definition
            elif properties[key]['type'] == "array" and "items" in properties[key] and "$ref" in properties[key]['items']:
                definition = properties[key]['items']['$ref'].replace("#/definitions/", "")
                if definition in definitions:
                    properties[key]['items'] = definitions[definition]

    return properties



"""
Loads include files in yaml.
"""

import yaml
import os.path
from .exceptions import FatalException


class Loader(yaml.Loader):

    def __init__(self, stream):

        self._root = os.path.split(stream.name)[0]

        super(Loader, self).__init__(stream)

    def include(self, node):

        filename = os.path.join(self._root, self.construct_scalar(node))

        extension = filename.split(".")[-1]

        with open(filename, 'r') as f:
            if extension in ["yaml", "raml", "yml", "json"]:  # defined by raml 1.0 spec
                return yaml.load(f, Loader)
            else:
                return f.read()

    def template(self, node):

        filename = os.path.join(self._root, self.construct_scalar(node))
        if os.path.isfile(filename):
            with open(filename, 'r') as f:
                return f.read()
        else:
            raise FatalException("Could not find %s" % filename)

Loader.add_constructor('!include', Loader.include)
Loader.add_constructor('!template', Loader.template)



import logging
import yaml

from django.urls import re_path

from .yaml_include_loader import Loader
from .validation import Endpoint, Action

logger = logging.getLogger(__name__)


def raml_url_patterns(raml_filepath, function_map):
    """
    creates url patterns that match the endpoints in the raml file, so can be quickly inserted into django urls.
    Note these
    :param raml_filepath: the path to the raml file (not a file pointer)
    :param function_map: a dictionary of urls to functions for mapping
    :return:
    """

    # This function will run in three phases:
    # 1) Load the raml (as a yaml document)
    # 2) Parse the raml into nodes that represent 'endpoints'
    # 3) Convert endpoints into a url structure

    # migrating from pyraml: file handling now has to be done by us
    # worry about streaming files in future version (for VERY BIG raml?)
    f = open(raml_filepath)
    tree = yaml.load(f, Loader=Loader)  # This loader has the !include directive
    f.close()

    # The resource map is the found nodes

    patterns = []
    to_look_at = [
        {
            "node": tree,
            "path": ""
        }
    ]

    # FIXME: get baseuri, and default media types out here

    defaults = {
        "content_type": "application/json",
    }

    for item in to_look_at:
        _parse_child(item, patterns, to_look_at, function_map, defaults)

    return patterns


def _parse_child(resource, patterns, to_look_at, function_map, defaults):

    node = resource['node']
    path = resource['path']
    local_endpoint = None

    for k in node:
        if k.startswith("/"):
            item = {
                "node": node[k],
                "path": "%s%s" % (path, k)
            }

            to_look_at.append(item)

        else:
            # attribute not subpath
            if path.startswith("/"):
                path = path[1:]

            if k in ("get", "post", "put", "delete", "patch"):
                act = node[k]

                if not local_endpoint:
                    local_endpoint = Endpoint(path)

                # look for a 200.body.{{content-type}}
                # and a 200.body.{{content-type}}.example

                a = Action()
                a.resp_content_type = defaults["content_type"]

                # FIXME: at some point allow a construct for multi-methods
                if path in function_map:
                    # Check for new style or old style definitions
                    if type(function_map[path]) is dict:
                        if "function" in function_map[path]:
                            # add the target function
                            a.target = function_map[path]["function"]

                        if "regex" in function_map[path]:
                            # Add dynamic value regex if present
                            local_endpoint.parse_regex(function_map[path]["regex"])
                    else:
                        # Deprecated! Ramlwrap < 2.0 compatibility
                        # I am not completely sure this is always desirable to fix though?
                        logger.warn("The function map for [%s] is not the 2.0 and above object - style. Please fix as this will be depricated in newer versions of RamlWrap (the fix is a simple copy/paste change to your code layout)" % path)
                        a.target = function_map[path]

                else:
                    # The path is not in a function map, check if it is a dynamic url as this will cause errors later
                    if "{" in path:
                        logger.error("Url: [%s] appears to have a dynamic component but there is no function map for it. You must define the regex in the function map to prevent errors" % path)

                if 'body' in act:
                    # if body, look for content type
                    # FIXME: this may be a bug requring a try/catch - need more real world example ramls
                    request_options = {}
                    request_content_type_options = []
                    # For each body defined in the raml, store the content type and schema if present
                    for i in act["body"].items():
                        request_content_type_options.append(i[0])
                        request_options[i[0]] = {"schema": None}

                        if len(i) > 1 and i[1]:
                            if "schema" in i[1]:
                                request_options[i[0]] = {"schema": i[1]["schema"]}

                    a.request_options = request_options
                    a.request_content_type_options = request_content_type_options


                # These horrendous if blocks are to get around none type errors when the tree
                # is not fully built out.

                if 'responses' in act and act['responses']:
                    for status_code in act['responses']:
                        # this is a response that we care about:
                        if status_code == 200:
                            two_hundred = act['responses'][200]
                            for resp_attr in two_hundred:
                                if resp_attr == "body":
                                    # not sure if this can fail and be valid raml?
                                    a.resp_content_type = next(iter(two_hundred['body']))
                                    if two_hundred['body'][a.resp_content_type]:
                                        if "example" in two_hundred['body'][a.resp_content_type]:
                                            a.example = two_hundred['body'][a.resp_content_type]['example']
                                        elif "examples" in two_hundred['body'][a.resp_content_type] and \
                                                two_hundred['body'][a.resp_content_type]['examples']:
                                            # If multiple examples in raml, just return the first one
                                            examples = two_hundred['body'][a.resp_content_type]['examples'].values()
                                            value_iterator = iter(examples)
                                            a.example = next(value_iterator)

                if "queryParameters" in act and act["queryParameters"]:
                    # FIXME: does this help in the query parameterising?
                    # For filling out a.queryparameterchecks
                    a.query_parameter_checks = act['queryParameters']

                local_endpoint.add_action(k.upper(), a)

    if local_endpoint:
        # strip leading
        if local_endpoint.url.startswith("/"):
            url_to_use = local_endpoint.url[1:]
        else:
            url_to_use = local_endpoint.url

        patterns.append(re_path("^%s$" % url_to_use, local_endpoint.serve))
